#include <jni.h>
#include <jvmti.h>
#include <stdlib.h>
#include <string.h>

#include "nbb.h"
#include "utils.h"
#include "async.h"
#include "ShadowData.h"

static jvmtiEnv * jvmti;

JNIEXPORT jint JNICALL heap_callback(jlong class_tag, jlong size,
		jlong * tag_ptr, jint length, void * user_data);

JNIEXPORT jobject JNICALL Java_ch_usi_inf_sape_ld_LeakDetector_getShadowState(
		JNIEnv * jniEnv, jclass ignored, jobject this) {
	jlong gottenTag;
	(*jvmti)->GetTag(jvmti, this, &gottenTag);
	return (jobject) gottenTag;
}

JNIEXPORT void JNICALL Java_ch_usi_inf_sape_ld_LeakDetector_setShadowState(
		JNIEnv * jniEnv, jclass ignored, jobject this,
		jobject shadowData) {
	// TODO matej Debug flag should enable this
//	jlong gottenTag;
//	(*jvmti)->GetTag(jvmti, this, &gottenTag);
//	if (gottenTag != 0) {
//		fprintf(stderr, "ERROR: please do not re-set tags\n");
//		abort();
//	}

	const jobject ref = (*jniEnv)->NewGlobalRef(jniEnv, shadowData);
	if (ref == NULL) {
		fprintf(stderr, "ERROR: could not allocate global reference\n");
		abort();
	}
	(*jvmti)->SetTag(jvmti, this, (jlong) ref);
}

JNIEXPORT void JNICALL Java_ch_usi_inf_sape_ld_LeakDetector_startDump(
		JNIEnv * jniEnv, jclass ignored) {
	jvmtiHeapCallbacks callbacks;
	(void) memset(&callbacks, 0, sizeof(callbacks));
	callbacks.heap_iteration_callback = &heap_callback;
	const jvmtiError err = (*jvmti)->IterateThroughHeap(jvmti,
			JVMTI_HEAP_FILTER_UNTAGGED, NULL, &callbacks, NULL);
	CHECK_JVMTI_ERROR(jvmti, err);
}

JNIEXPORT void JNICALL gc_callback(jvmtiEnv * env, jlong tag) {
nbb_write(&gc_buffer, tag);
}

JNIEXPORT jint JNICALL heap_callback(jlong class_tag, jlong size,
	jlong * tag_ptr, jint length, void * user_data) {
nbb_write(&heap_buffer, *tag_ptr);
return JVMTI_VISIT_OBJECTS;
}

JNIEXPORT jint JNICALL JNI_OnLoad(JavaVM *vm, void *reserved) {
jvmti = NULL;
const jint err = (*vm)->GetEnv(vm, (void **) &jvmti, JVMTI_VERSION);

if (err != JNI_OK) {
	fprintf(stderr, "ERROR: GetEnv failed, error=%d\n", err);
	abort();
}

set_caps(jvmti);
set_callbacks(jvmti, &gc_callback);

run_threads(vm);

return JNI_VERSION_1_6;
}
