#include <stdio.h>
#include <stdlib.h>
#include <jni.h>
#include <sys/time.h>

#include "nbb.h"

void nbb_init(const nbbref nbb) {
	one_buffer_ref buffer = (one_buffer_ref) malloc(sizeof(one_buffer_t));

	buffer->next = NULL;
	buffer->start = 0;

	nbb->first = buffer;
	nbb->last = buffer;

	pthread_mutex_init(&nbb->lock, NULL );
	pthread_cond_init(&nbb->full, NULL );
}

void nbb_write(const nbbref nbb, const jlong elem) {
	int res = pthread_mutex_lock(&nbb->lock);
	if (res != 0) {
		printf("WRITE: pthread_mutex_lock failed with errno %d\n", res);
		exit(1);
	}
	nbb->last->elems[nbb->last->start] = elem;
	nbb->last->start++;
	if (nbb->last->start >= NBB_BUFFER_SIZE) {
		one_buffer_ref buffer = (one_buffer_ref) malloc(sizeof(one_buffer_t));

		buffer->next = NULL;
		buffer->start = 0;

		nbb->last->next = buffer;
		nbb->last = buffer;

		res = pthread_cond_signal(&nbb->full);
		if (res != 0) {
			printf("WRITE: pthread_cond_signal failed with errno %d\n", res);
			exit(1);
		}
	}
	res = pthread_mutex_unlock(&nbb->lock);
	if (res != 0) {
		printf("WRITE: pthread_mutex_unlock failed with errno %d\n", res);
		exit(1);
	}
}

void nbb_read_buffer(const nbbref nbb, jlong * buffer) {
	int res = pthread_mutex_lock(&nbb->lock);
	if (res != 0) {
		printf("READ: pthread_mutex_lock failed with errno %d\n", res);
		exit(1);
	}
	struct timespec ts;
	struct timeval tp;
	while (nbb->first == nbb->last) {
		gettimeofday(&tp, NULL );
		ts.tv_sec = tp.tv_sec;
		ts.tv_nsec = (tp.tv_usec + 1) * 1000;
		pthread_cond_timedwait(&nbb->full, &nbb->lock, &ts);
	}
	int i;
	for (i = 0; i < NBB_BUFFER_SIZE; i++) {
		buffer[i] = nbb->first->elems[i];
	}
	one_buffer_ref first = nbb->first;
	nbb->first = nbb->first->next;
	free(first);
	res = pthread_mutex_unlock(&nbb->lock);
	if (res != 0) {
		printf("READ: pthread_mutex_unlock failed with errno %d\n", res);
		exit(1);
	}
}
