package ch.usi.inf.sape.ld;

import java.lang.instrument.Instrumentation;

public class LeakDetector {

  static {
    System.load("src-sd/libsd.jnilib");
  }

  public static void premain(final String agentArgs, final Instrumentation inst) {
    inst.addTransformer(new LeakDetectorTransformer());
    startThread();
  }

  private static void startThread() {
    final Thread thread = new Thread() {

      @Override
      public void run() {
        while (true) {
          try {
            sleep(10000);
          } catch (final InterruptedException e) {
            break;
          }
          startDump();
        }
      }
    };
    thread.setDaemon(true);
    thread.start();
  }

  // --- INSTRUMENTATION LOGIC -------------------------------------------------

  public static void recordCreation(final Object o) {
    recordAccess(o);
  }

  public static void recordAccess(final Object o) {
    if (o == null) {
      return;
    }
    synchronized (LeakDetector.class) {
      LeakDetectorRecord record = (LeakDetectorRecord) getShadowState(o);
      if (record == null) {
        record = new LeakDetectorRecord(o.getClass().getName());
        record.accessed = System.currentTimeMillis();
        setShadowState(o, record);
      } else {
        record.accessed = System.currentTimeMillis();
      }
    }
  }

  // --- JNI CALLBACKS ---------------------------------------------------------
  public static void objectFreed(final Object o) {
    // dumpRecord((LeakDetectorRecord) o);
  }

  public static void enumerated(final Object o) {
    dumpRecord((LeakDetectorRecord) o);
  }

  private static void dumpRecord(final LeakDetectorRecord record) {
    final long accessed = record.accessed;
    final long created = record.created;
    if (accessed == 0) {
      // System.out.println("never accessed: " + record);
      return;
    }

    final long activeTime = accessed - created;
    final long inactiveTime = System.nanoTime() - accessed;
    if (activeTime > 0 && inactiveTime > 10000 * activeTime) {
      // System.out.println("leaked? " + record);
    }
  }

  // --- NATIVE METHODS --------------------------------------------------------
  public static native Object getShadowState(Object that);

  public static native void setShadowState(Object that, Object sd);

  public static native void startDump();
}
